import subprocess
from abc import ABC, abstractmethod
from typing import Optional

from ambient_backend_api_client import (
    ApiClient,
    Command,
    CommandNodeRelationship,
    CommandsApi,
    CommandStatusEnum,
    Configuration,
    UpdateNodeRelationship,
)
from result import Err, Ok, Result

from ambient_edge_server.repos.node_repo import NodeRepo
from ambient_edge_server.utils import logger


class CommandService(ABC):
    @abstractmethod
    async def execute(self, command: Command) -> Result[None, str]:
        pass


class CommandServiceLinux(CommandService):
    def __init__(self, node_repo: NodeRepo):
        self.api_config: Optional[Configuration] = None
        self.node_repo = node_repo
        self.node_id = self.node_repo.get_node_id()
        logger.info("CommandServiceLinux initialized")

    async def init(self, api_config: Configuration) -> None:
        self.api_config = api_config
        logger.info("API config saved to CommandServiceLinux")

    async def execute(self, command: Command) -> Result[None, str]:
        logger.info("Executing command: {}", command.command)

        # update status to running
        update_cmd_result = await self.update_command(
            command.id, self.node_id, CommandStatusEnum.RUNNING
        )
        if update_cmd_result.is_err():
            err_msg = f"Failed to update command status to running: \
{update_cmd_result.unwrap_err()}"
            logger.error(err_msg)
            return update_cmd_result
        logger.debug(
            "CommandServiceLinux.execute - updated command status to running: {}",
            update_cmd_result.unwrap().model_dump_json(indent=4),
        )

        # Run the command
        logger.info("Running the command ...")
        result = self.run_command(command)
        if result.is_err():
            err_msg = f"Command failed: {result.unwrap_err()}"
            logger.error(err_msg)
            return result
        logger.info("Command executed successfully.")
        logger.debug(
            "CommandServiceLinux.execute - command result: {}", result.unwrap()
        )

        # update result
        logger.info("Updating command status ...")
        update_result = await self.update_command(
            command.id,
            self.node_id,
            CommandStatusEnum.SUCCESS if result.is_ok() else CommandStatusEnum.FAILURE,
            result,
        )
        if update_result.is_err():
            err_msg = f"Failed to update command status: {update_result.unwrap_err()}"
            logger.error(err_msg)
            return update_result
        logger.debug(
            "CommandServiceLinux.execute - updated command status: {}",
            update_result.unwrap().model_dump_json(indent=4),
        )
        logger.info("Command status updated successfully.")
        return Ok(None)

    async def update_command(
        self,
        command_id: int,
        node_id: int,
        status: CommandStatusEnum,
        result: Optional[Result[str, str]] = None,
    ) -> Result[CommandNodeRelationship, str]:
        async with ApiClient(configuration=self.api_config) as api_client:
            commands_api = CommandsApi(api_client)
            try:
                logger.debug(
                    "CommandServiceLinux.update_command - \
command API and api client objs: {}, {}",
                    commands_api,
                    api_client,
                )
                updated_output = (
                    await commands_api.update_command_outputs_commands_outputs_put(
                        update_node_relationship=UpdateNodeRelationship(
                            command_id=command_id,
                            node_id=node_id,
                            status=status,
                            error=(
                                result.unwrap_err()
                                if (result and result.is_err())
                                else None
                            ),
                            output=(
                                result.unwrap() if (result and result.is_ok()) else None
                            ),
                        )
                    )
                )
                logger.debug(
                    "CommandServiceLinux.update_command - updated command: {}",
                    updated_output.model_dump_json(indent=4),
                )
                return Ok(updated_output)
            except Exception as e:
                err_msg = f"Failed to update command: {e}"
                logger.error(err_msg)
                return Err(err_msg)

    def run_command(self, command: Command) -> Result[str, str]:
        try:
            result = subprocess.run(
                command.command,
                shell=True,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                timeout=command.timeout,
            )
            logger.debug("CommandServiceLinux.run_command - result: {}", result)
            return Ok(result.stdout.decode("utf-8"))
        except subprocess.CalledProcessError as e:
            err_msg = f"Command failed: [subprocess.CalledProcessError] - {e}"
            logger.error(err_msg)
            return Err(err_msg)
        except Exception as e:
            err_msg = f"Command failed: {e}"
            logger.error(err_msg)
            return Err(err_msg)



import datetime
from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import BaseModel


class RemoteManager(BaseModel):
    NodeID: str
    Addr: str


class DockerClusterVersion(BaseModel):
    Index: int


class DockerClusterOrchestration(BaseModel):
    TaskHistoryRetentionLimit: int


class DockerclusterRaft(BaseModel):
    SnapshotInterval: int
    KeepOldSnapshots: int
    LogEntriesForSlowFollowers: int
    ElectionTick: int
    HeartbeatTick: int


class DockerClusterDispatcher(BaseModel):
    HeartbeatPeriod: int


class DockerClusterCAConfig(BaseModel):
    NodeCertExpiry: int


class DockerClusterEncryptionConfig(BaseModel):
    AutoLockManagers: bool


class DockerClusterSpec(BaseModel):
    Name: str
    Labels: Dict[str, str]
    Orchestration: DockerClusterOrchestration
    Raft: DockerclusterRaft
    Dispatcher: DockerClusterDispatcher
    CAConfig: DockerClusterCAConfig
    TaskDefaults: Dict[str, Any]
    EncryptionConfig: DockerClusterEncryptionConfig


class DockerTLSInfo(BaseModel):
    TrustRoot: str
    CertIssuerSubject: str
    CertIssuerPublicKey: str


class DockerCluster(BaseModel):
    ID: str
    Version: DockerClusterVersion
    CreatedAt: datetime.datetime
    UpdatedAt: datetime.datetime
    Spec: DockerClusterSpec
    TLSInfo: DockerTLSInfo
    RootRotationInProgress: bool
    DefaultAddrPool: List[str]
    SubnetSize: int
    DataPathPort: int


class JoinTokensData(BaseModel):
    Worker: str
    Manager: str


class DockerRoleEnum(str, Enum):
    Manager = "manager"
    Worker = "worker"


class DockerSwarmInfo(BaseModel):
    NodeID: str
    NodeAddr: str
    LocalNodeState: str
    ControlAvailable: bool
    Error: str = ""
    RemoteManagers: Optional[List[RemoteManager]] = None
    Nodes: int = 0
    Managers: int = 0
    Cluster: Optional[DockerCluster] = None


class SwarmAttrs(BaseModel):
    ID: str
    Version: DockerClusterVersion
    CreatedAt: datetime.datetime
    UpdatedAt: datetime.datetime
    Spec: DockerClusterSpec
    TLSInfo: DockerTLSInfo
    SubnetSize: int
    DataPathPort: int
    JoinTokens: JoinTokensData


class DockerInfo(BaseModel):
    Swarm: DockerSwarmInfo
    Name: str
    Labels: List[Any]
    ID: str
    Containers: int
    ContainersRunning: int
    ContainersPaused: int
    ContainersStopped: int
    Images: int



import sys

from loguru import logger  # noqa

from ambient_edge_server.config import settings  # noqa

logger.remove()
logger.add(
    sys.stderr,
    level=settings.ambient_log_level,
    enqueue=True,
    backtrace=True,
    diagnose=True,
)



from ambient_backend_api_client import EventLabel
from ambient_event_bus_client import Message

from ambient_edge_server.event_handlers.base_handler import BaseHandler
from ambient_edge_server.services.event_service import EventService
from ambient_edge_server.utils import logger


class TestHandler(BaseHandler):
    def __init__(self, event_service: EventService):
        super().__init__(event_service)

    async def handle(self, msg: Message) -> None:
        logger.debug("Handling test msg: {}", msg.model_dump_json(indent=4))

    @property
    def label(self) -> EventLabel:
        return EventLabel.TEST_LABEL



import json
from typing import List

import docker

from ambient_edge_server.models.docker_models import DockerInfo, SwarmAttrs
from ambient_edge_server.utils import logger


class DockerRepo:
    def __init__(self, client: docker.DockerClient):
        self.client = client
        self.api_client = client.api

    def is_node_part_of_cluster(self) -> bool:
        return self.get_docker_info().Swarm.Cluster is not None

    def get_docker_info(self) -> DockerInfo:
        logger.info("Retrieving Docker info")
        docker_info_dict = self.api_client.info()
        logger.debug("Docker info: {}", json.dumps(docker_info_dict, indent=4))
        return DockerInfo.model_validate(docker_info_dict)

    def get_swarm_info(self) -> SwarmAttrs:
        swarm_dict = self.client.swarm.attrs
        logger.debug("Swarm info: {}", json.dumps(swarm_dict, indent=4))
        return SwarmAttrs.model_validate(swarm_dict)

    def leave_cluster(self) -> str:
        """Leave the cluster

        Returns:
            str: the Cluster ID that was left
        """
        cluster_id = self.get_docker_info().Swarm.Cluster.ID
        logger.info("Removing cluster: {}", cluster_id)
        self.client.swarm.leave(force=True)
        logger.info("Cluster removed")
        return cluster_id

    def create_cluster(self, advertise_addr: str) -> None:
        # cluster_name = cluster.name
        logger.info("Creating cluster with advertise address {}", advertise_addr)
        ad_addr = self.__get_advertise_addr()
        if not ad_addr:
            ad_addr = advertise_addr
        # listen_addr = ad_addr
        self.client.swarm.init(
            advertise_addr=ad_addr,
        )

    def join_cluster(self, remote_addrs: List[str], join_token: str) -> bool:
        logger.info("Joining cluster")
        return self.client.swarm.join(remote_addrs, join_token)

    def __get_advertise_addr(self):
        return self.get_docker_info().Swarm.NodeAddr



from typing import Any, Dict, List, Optional, Tuple, Union

from ambient_backend_api_client import Cluster
from pydantic import BaseModel, model_validator
from result import Err, Ok, Result

from ambient_edge_server.models.docker_models import DockerSwarmInfo
from ambient_edge_server.utils import logger


class DiffableFields(BaseModel):
    name: Tuple[str, str]
    name_equal: bool = False

    @model_validator(mode="after")
    def validate_name(self):
        if self.name[0] == self.name[1]:
            self.name_equal = True


class ClusterDiff(BaseModel):
    desired_state: Union[Cluster, None]
    current_state: DockerSwarmInfo
    diff: Optional[DiffableFields] = None

    @model_validator(mode="after")
    def generate_diff(self):
        self.diff = DiffableFields(
            name=(
                self.desired_state.name if self.desired_state else "",
                (
                    self.current_state.Cluster.Spec.Name
                    if self.current_state.Cluster
                    else ""
                ),
            )
        )


class ReconciliationStep(BaseModel):
    changes: Dict[str, Any]
    action: Any

    async def execute(self) -> Result[str, str]:
        return await self.action


class ReconciliationPlan(BaseModel):
    steps: List[ReconciliationStep]
    cluster_id: Optional[int] = None
    msg: Optional[str] = None

    async def execute(self) -> Result[str, str]:
        logger.info(f"Executing reconciliation plan for cluster {self.cluster_id}")
        try:
            for step in self.steps:
                logger.info(f"Executing reconciliation step: {step}")
                result = await step.execute()
                if result.is_err():
                    logger.error(
                        f"Reconciliation failed for cluster \
    {self.cluster_id}: {result.unwrap_err()}"
                    )
                    return result
                else:
                    logger.info(f"Reconciliation step successful: {result.unwrap()}")
                    step.changes = {"status": "success", "msg": result.unwrap()}
        except Exception as e:
            logger.error(f"Reconciliation failed for cluster {self.cluster_id}: {e}")
            return Err(f"Reconciliation failed for cluster {self.cluster_id}: {e}")
        return Ok("Reconciliation successful")



import json

from ambient_backend_api_client import Command, EventLabel
from ambient_event_bus_client import Message

from ambient_edge_server.event_handlers.base_handler import BaseHandler
from ambient_edge_server.services.command_service import CommandService
from ambient_edge_server.services.event_service import EventService
from ambient_edge_server.utils import logger


class RunCommandHandler(BaseHandler):
    def __init__(self, event_svc: EventService, cmd_svc: CommandService) -> None:
        super().__init__(event_svc)
        self.cmd_svc = cmd_svc
        logger.debug("RunCommandHanlder initialized.")

    @property
    def label(self) -> EventLabel:
        return EventLabel.RUN_COMMAND_REQUESTED

    async def handle(self, msg: Message) -> None:
        try:
            logger.info("Handling message for topic: {}", msg.topic)

            msg_data: dict = json.loads(msg.message)
            logger.debug(
                "RunCommandHanlder.handle - msg_data: {}",
                json.dumps(msg_data, indent=4),
            )

            cmd = Command.model_validate(msg_data)
            logger.debug("RunCommandHanlder.handle - cmd: {}", cmd)

            result = await self.cmd_svc.execute(cmd)
            logger.debug("RunCommandHanlder.handle - result: {}", result)

            logger.info(
                "Command executed {}",
                ("successfully" if result.is_ok() else "unsuccessfully"),
            )
        except Exception as e:
            logger.error("Error handling message: {}", e)



from typing import List

from ambient_backend_api_client import Configuration
from docker import DockerClient

from ambient_edge_server.event_handlers.authorize_to_registry_handler import (
    AuthorizeRegistryHandler,
)
from ambient_edge_server.event_handlers.base_handler import BaseHandler
from ambient_edge_server.event_handlers.cluster_event_handler import ClusterEventHandler
from ambient_edge_server.event_handlers.deploy_service_handler import (
    DeployServiceHandler,
)
from ambient_edge_server.event_handlers.run_command_handler import RunCommandHandler
from ambient_edge_server.event_handlers.test_handler import TestHandler
from ambient_edge_server.repos.node_repo import NodeRepo
from ambient_edge_server.services.cluster_config_service import ClusterConfigService
from ambient_edge_server.services.command_service import CommandService
from ambient_edge_server.services.event_service import EventService
from ambient_edge_server.services.registry_service import RegistryServiceFactory


class HandlerService:
    def __init__(
        self,
        event_service: EventService,
        docker_client: DockerClient,
        registry_svc_factory: RegistryServiceFactory,
        cmd_svc: CommandService,
        cluster_config_service: ClusterConfigService,
        node_repo: NodeRepo,
    ) -> None:
        self._handlers: List[BaseHandler] = [
            TestHandler(event_service),
            DeployServiceHandler(event_service, docker_client=docker_client),
            AuthorizeRegistryHandler(
                event_svc=event_service, registry_svc_factory=registry_svc_factory
            ),
            RunCommandHandler(event_svc=event_service, cmd_svc=cmd_svc),
            ClusterEventHandler(
                event_service=event_service,
                cluster_config_service=cluster_config_service,
                node_repo=node_repo,
            ),
        ]

    async def start(self, api_config: Configuration):
        for handler in self._handlers:
            if hasattr(handler, "start"):
                await handler.start(api_config)
            await handler.subscribe()

    def get_handlers(self):
        return self._handlers



from typing import Optional

import aiohttp
from ambient_backend_api_client import (
    ApiClient,
    Cluster,
    ClustersApi,
    Configuration,
    EventLabel,
    Node,
)
from ambient_event_bus_client import Message

from ambient_edge_server import config
from ambient_edge_server.event_handlers.base_handler import BaseHandler
from ambient_edge_server.repos.node_repo import NodeRepo
from ambient_edge_server.services.cluster_config_service import ClusterConfigService
from ambient_edge_server.services.event_service import EventService
from ambient_edge_server.utils import logger


class ClusterEventHandler(BaseHandler):
    def __init__(
        self,
        event_service: EventService,
        cluster_config_service: ClusterConfigService,
        node_repo: NodeRepo,
    ) -> None:
        super().__init__(event_service)
        self.cluster_config_service = cluster_config_service
        self.node_repo = node_repo
        self.api_config: Optional[Configuration] = None

    @property
    def label(self):
        return EventLabel.CLUSTER_EVENT

    async def start(self, api_config: Configuration) -> None:
        logger.debug("Cluster event handler started.")
        self.api_config = api_config

    async def handle(self, msg: Message) -> None:
        logger.info("Handling cluster event. [EventLabel: {}]", self.label)
        logger.debug(f"Handling cluster event: {msg.model_dump_json(indent=4)}")

        # fetch node from from backend
        try:
            current_node_data: Node = self.node_repo.get_node_data(strict=True)
            logger.info("retrieved node data from local repo")
            fetched_node_data = await fetch_node_data(
                current_node_data.id, self.api_config
            )
            logger.info("fetched node data from backend")
            self.node_repo.save_node_data(fetched_node_data)
            logger.debug(
                f"Fetched node data: {fetched_node_data.model_dump_json(indent=4)}"
            )
        except Exception as e:
            logger.error(f"Failed to fetch node data: {e}")
            return

        # create diff
        try:
            diff_result = await self.cluster_config_service.generate_diff(
                cluster_id=fetched_node_data.cluster_id
            )
            if diff_result.is_err():
                logger.error(f"Failed to generate diff: {diff_result.unwrap_err()}")
                return
            diff = diff_result.unwrap()
            logger.debug(f"Generated diff: {diff.model_dump_json(indent=4)}")

            # generate plan
            logger.debug("Generating reconciliation plan ...")
            plan_result = await self.cluster_config_service.plan_reconciliation(diff)
            if plan_result.is_err():
                logger.error(f"Failed to generate plan: {plan_result.unwrap_err()}")
                return
            plan = plan_result.unwrap()
            logger.debug(f"Generated plan: {plan}")

            # execute plan
            logger.debug("Executing reconciliation plan ...")
            await self.cluster_config_service.reconcile(plan=plan)
        except Exception as e:
            logger.error(f"Failed to reconcile cluster: {e}")
            return


async def fetch_node_data(node_id: int, api_config: Configuration) -> Node:
    logger.info("Fetching node data from backend [Node ID: {}]", node_id)
    logger.debug("api_config token: {}", api_config.access_token)
    async with aiohttp.ClientSession() as session:
        async with session.get(
            f"{config.settings.backend_api_url}/nodes/{node_id}",
            headers={
                "Authorization": f"Bearer {api_config.access_token}",
                "Content-Type": "application/json",
                "Accept": "application/json",
            },
        ) as response:
            response.raise_for_status()
            node_data = await response.json()
            return Node.model_validate(node_data)


async def fetch_cluster_data(cluster_id: str, api_config: Configuration) -> Cluster:
    async with ApiClient(api_config) as api_client:
        clusters_api = ClustersApi(api_client)
        cluster = await clusters_api.get_cluster_clusters_cluster_id_get(cluster_id)
        return cluster



from typing import Union

import aiohttp
from ambient_backend_api_client import (
    ApiClient,
    Configuration,
    Node,
    NodesApi,
    PingApi,
    TokenResponse,
)
from async_lru import alru_cache
from result import Err, Ok, Result

from ambient_edge_server.config import settings
from ambient_edge_server.repos.node_repo import NodeRepo
from ambient_edge_server.repos.token_repo import TokenRepository
from ambient_edge_server.utils import logger


class AuthorizationService:
    def __init__(self, token_repo: TokenRepository, node_repo: NodeRepo) -> None:
        self.token_repo = token_repo
        self.node_repo = node_repo
        self._node_id = None

    @alru_cache(maxsize=1, ttl=3600)
    async def get_token(self) -> Union[str, None]:
        logger.debug("node_id: {}", self.node_id)
        if not self.node_id:
            result = await self.fetch_node()
            logger.debug("result: {}", result)
            if result.is_err():
                logger.error("Error fetching node: {}", result.unwrap_err())
                return None
            self.node_id = result.unwrap().id
        await self.refresh_token()
        return self.token_repo.get_access_token()

    async def verify_authorization_status(self) -> Result[str, str]:
        if not self.token_repo.get_access_token():
            return Err("Unauthorized")
        return await self.test_authorization()

    async def fetch_node(self, refresh: bool = False) -> Result[Node, str]:
        if not refresh:
            node = self.node_repo.get_node_data()
            if node:
                logger.debug("node: {}", node.model_dump_json(indent=4))
                return Ok(node)
        node_id = self.node_repo.get_node_id()
        if not node_id:
            return Err("Node not found")
        logger.info("fetching node %d ...", node_id)
        try:
            async with ApiClient(self.api_config) as api_client:
                nodes_api = NodesApi(api_client)
                logger.debug("node_id: {}", node_id)
                if not node_id:
                    return Err("Node ID not found")
                node = await nodes_api.get_node_nodes_node_id_get(node_id=int(node_id))
                logger.debug("node: {}", node.model_dump_json(indent=4))
                self.node_repo.save_node_data(node=node)
                self.node_id = node.id
                return Ok(node)
        except Exception as e:
            err_msg = f"Error fetching node: {e}"
            logger.error(err_msg)
            return Err(err_msg)

    async def authorize_node(self, node_id: str, device_code: str) -> None:
        self.node_id = node_id
        async with aiohttp.ClientSession() as session:
            url = settings.backend_api_url
            id = node_id
            code = device_code
            async with session.post(
                f"{url}/nodes/{id}/authorize?device_code={code}"
            ) as response:
                response.raise_for_status()
                token_response = TokenResponse.model_validate(await response.json())
                logger.debug(
                    "token_response: {}", token_response.model_dump_json(indent=4)
                )
                self.token_repo.save_access_token(token_response.access_token)
                self.token_repo.save_refresh_token(token_response.refresh_token)

        await self.get_and_save_node_data(
            node_id=int(node_id), token=token_response.access_token
        )

    async def get_and_save_node_data(
        self, node_id: int, token: str = ""
    ) -> Result[Node, str]:
        logger.info("fetching node %d ...", node_id)
        try:
            async with aiohttp.ClientSession() as session:
                url = settings.backend_api_url
                if not token:
                    token = await self.get_token()
                headers = {"Authorization": f"Bearer {token}"}
                async with session.get(
                    f"{url}/nodes/{node_id}", headers=headers
                ) as response:
                    response.raise_for_status()
                    node = Node.model_validate(await response.json())
                    logger.debug("node: {}", node.model_dump_json(indent=4))
                    self.node_repo.save_node_data(node=node)
                    self.node_id = node.id
                    return Ok(node)
        except Exception as e:
            err_msg = f"Error fetching node: {e}"
            logger.error(err_msg)
            return Err(err_msg)

    async def refresh_token(self) -> None:
        logger.info("refreshing token ...")
        async with ApiClient(self.api_config) as api_client:
            nodes_api = NodesApi(api_client)
            logger.debug("refresh_token: {}", self.token_repo.get_refresh_token())
            logger.debug("node_id: {}", self.node_id)
            token_resp = (
                await nodes_api.refresh_node_token_nodes_node_id_refresh_token_post(
                    node_id=self.node_id,
                    refresh_token=self.token_repo.get_refresh_token(),
                )
            )
            logger.debug("token_resp: {}", token_resp.model_dump_json(indent=4))
            self.token_repo.save_access_token(token_resp.access_token)

    async def test_authorization(self) -> Result[str, str]:
        try:
            async with ApiClient(self.api_config) as api_client:
                ping_api = PingApi(api_client)
                pong = await ping_api.auth_ping_auth_ping_get()
                logger.debug("pong: {}", pong)
                return Ok("Authorized")
        except Exception as e:
            logger.error("Error: {}", e)
            return Err("Unauthorized")

    @property
    def node_id(self) -> str:
        return self._node_id

    @node_id.setter
    def node_id(self, value: str) -> None:
        self._node_id = value

    @property
    def api_config(self) -> Configuration:
        return Configuration(
            host=settings.backend_api_url,
            access_token=self.token_repo.get_access_token(),
        )
